#include "lists.h"

/* Necessary Initializations */
LISTPTR LIST__Initialize ()
{
	LISTPTR list = (LISTPTR) malloc (sizeof (LIST));

	/* NIL's Initialization */
	list->NIL = (LIST_NODEPTR) malloc (sizeof (LIST_NODE));
	list->NIL->key_code = NULL_KEY;
	list->NIL->lesson_node = NULL;
	list->NIL->next = list->NIL;
	list->NIL->previous = list->NIL;

	/* HEAD's Initialization */
	list->HEAD = list->NIL;

	/* Nothing is yet inserted */
	list->NUMBER_OF_ELEMENTS = 0;

	return list;	// Return new list.
}


/* List's Search Operation */
LIST_NODEPTR LIST__Search (LISTPTR list, KEY_TYPE key)
{
	LIST_NODEPTR current = list->HEAD;

	while ((current != list->NIL) && (current->key_code != key))
		current = current->next;	// Procceed to next element in list.

	return current;
}


/* Insert node at front - Do not allow duplicate insertions according to key */
int LIST__Insert_node_at_front (LISTPTR list, KEY_TYPE lesson_code, TREE_NODEPTR lesson_tree_node)
{
	LIST_NODEPTR new_node = (LIST_NODEPTR) malloc (sizeof (LIST_NODE));
	LIST_NODEPTR search_node;

	search_node = LIST__Search (list, lesson_code);

	if (search_node != list->NIL) {
		free (new_node);	// Deallocate memory.

		if (search_node->lesson_node == lesson_tree_node)
			return 0;		// Indicate key was previously inserted in list.
		else
			return -1;		// Indicate error with tree's nodes.
	}

	/* Node's Initializations */
	new_node->key_code = lesson_code;
	new_node->lesson_node = lesson_tree_node;

	/* List's Modifications */
	new_node->next = list->HEAD;
	
	if (list->HEAD != list->NIL)
		(list->HEAD)->previous = new_node;

	list->HEAD = new_node;
	new_node->previous = list->NIL;

	list->NUMBER_OF_ELEMENTS++;		// Increment counter;

	return 1;	// Indicate successful insertion.
}


/* Search List for a node with KEY == key and if found, delete it. */
int LIST__Delete_node_according_to_key (LISTPTR list, KEY_TYPE key)
{
	LIST_NODEPTR search_node;

	if (list->NUMBER_OF_ELEMENTS == 0)
		return -1;		// Indicate list is empty.

	search_node = LIST__Search (list, key);

	if (search_node == list->NIL)
		return 0;		// Node with corresponding key was not found.

	if (search_node->previous != list->NIL)
		(search_node->previous)->next = search_node->next;
	else
		list->HEAD = search_node->next;

	if (search_node->next != list->NIL)
		(search_node->next)->previous = search_node->previous;

	free (search_node);		// Deallocate memory.

	list->NUMBER_OF_ELEMENTS--;		// Decrement counter.

	return 1;	// Indicate successful deletion.
}


/* Destroy List */
int LIST__Destroy (LISTPTR list)
{
	LIST_NODEPTR current;
	int key;

	if (list->NUMBER_OF_ELEMENTS > 0) {
		current = list->HEAD;
		key = current->key_code;
		while (list->NUMBER_OF_ELEMENTS > 0) {
			current = list->HEAD;
			key = current->key_code;
			LIST__Delete_node_according_to_key (list, key);
		}
	}

	/* Deallocate remaining memory. */
	free (list->NIL);
	free (list);

	return 1;	// Indicate successful destruction.
}


/* Swap Nodes in List */
int LIST__Swap_nodes (LISTPTR List, LIST_NODEPTR node1, LIST_NODEPTR node2)
{
	TREE_NODEPTR temp_lesson_node;
	KEY_TYPE temp_key;

	if (node1 == List->NIL || node2 == List->NIL)
		return 0;	// Indicate no swaping occurred.

	temp_key = node2->key_code;
	temp_lesson_node = node2->lesson_node;

	/* Change contents in nodes */
	node2->key_code = node1->key_code;
	node2->lesson_node = node1->lesson_node;
	node1->key_code = temp_key;
	node1->lesson_node = temp_lesson_node;

	return 1;		// Indicate successful swaping.
}

/* Union of 2 lists */
/* The new list does not have duplicate copies of the same element */
LISTPTR LIST__Union (LISTPTR list1, LISTPTR list2)
{
	LISTPTR unified = LIST__Initialize ();
	LIST_NODEPTR running_node;

	running_node = list1->HEAD;
	while (running_node != list1->NIL) {
		LIST__Insert_node_at_front (unified, running_node->key_code, running_node->lesson_node);
		running_node = running_node->next;
	}

	running_node = list2->HEAD;
	while (running_node != list2->NIL) {
		LIST__Insert_node_at_front (unified, running_node->key_code, running_node->lesson_node);
		running_node = running_node->next;
	}

	return unified;
}


/* BubbleSort applied on Lists - DECREASING ORDER */
void LIST__BubbleSort (LISTPTR List)
{
	LIST_NODEPTR j_node;
	int N = List->NUMBER_OF_ELEMENTS;
	int i, j;


	if (List->NUMBER_OF_ELEMENTS > 1) {
		for (i = 2; i <= N; i++) {
			j_node = List->HEAD;
			while (j_node->next != List->NIL)
				j_node = j_node->next;
			for (j = N; j >= i; j--) {
				if (j_node->lesson_node->data->grade > j_node->previous->lesson_node->data->grade)
					LIST__Swap_nodes (List, j_node, j_node->previous);
				j_node = j_node->previous;
			}
		}
	}
}
